RHTableViewProvider
===================

UITableViews wired up in a few lines of code? Yes sir.

## Updates

* Updated for iOS 7 and Auto Layout
* Custom cell drawing has been removed.. for custom cells add all views as subviews of 'customView' on 'RHTableViewProviderCell' subclasses
* Pull to refresh removed.. please implement UIRefreshControl on your tableView as required

## What Does It Do?

This is a set of classes to strip the UITableView boilerplate code from your view controllers, wire them up quickly, easily create complex, custom tableview setups with interchangeable cell and section styles/content.

* Wires a UITableView up in your view controller with three lines of code
* Easily provide custom objects to each cell
* Interchange custom UITableViewCell classes
* Custom drawing of cells (drawRect) if needed
* Handles display of your custom 'empty state' view
* Core Data compatibility setting table content with an NSFetchRequest and NSManagedObjectContext
* Interchange custom section header/footer views, populate with dynamic data
* Plays nicely with editing tables, deleting rows etc

Beyond the above, UITableView obviously offers more functionality via further delegate methods etc which I've not really found the need to package for re-use as yet. If you need to get more control and hook into the further parts of UITableView just subclass an instance of RHTableViewProvider and drop in your own delegates and notifications there as you require.


## Sample Usage

    @interface ViewController ()

    @property (strong, nonatomic) RHTableViewProvider *provider;
    @property (strong, nonatomic) UITableView *tableView;

    @end

    @implementation ViewController

    - (void)viewDidLoad
    {
      [super viewDidLoad];
      
      // Setup your table view provider
      self.provider = [[RHTableViewProvider alloc] initWithTableView:self.tableView delegate:self customise:NO];
      
      // Populate the table view with some content
      [_provider setContent:@[@"One", @"Two", @"Three"] withSections:NO];
    }

    #pragma mark - RHTableViewProviderDelegate

    - (void)RHTableViewProvider:(RHTableViewProvider *)provider didSelectRowAtIndexPath:(NSIndexPath *)indexPath
    {
      id object = [provider objectAtIndexPath:indexPath];
      NSLog(@"Hello Object: %@", object);
    }

    @end

## Custom Cells

All of your cells will need to be a subclass of __RHTableViewProviderCell__ which includes lots of useful properties when working with more intricate tables, template methods. A simple example of how to start working with a custom cell is below, setting the height for this type of cell, the __populateWithObject:(id)object__ implemenation which is how all our cells get their content, which can of course be passing or storing a reference to any object you like. 

    @implementation MYCustomCell

    // Margins to customize for your cell to include in content height calculations later
    + (CGFloat)margins
    {
      return 10.0f;
    }

    // Default height for this cell
    + (CGFloat)height
    {
      return 44.0f;
    }

    // Should the cell expand when selected
    + (BOOL)expandable
    {
      return NO;
    }

    // If the cell should expand, this is the height it will expand to
    + (CGFloat)heightForSelectedState
    {
      return 88.0f;
    }

    // Any object you wish to pass into the cell to populate it with, to work with dynamic cell heights this object will need to conform to RHTableViewProviderContent protocol
    - (void)populateWithObject:(id <RHTableViewProviderContent>)anObject
    {
      Person *person = (Person *)anObject;
      _nameLabel = person.name;
    }

    @end

## TODOs

* Cell Asset Caching option
* Improve the dynamic cell height logic to simplify the setup and allow of cells to expand/collapse to dynamic heights
* Add some tests :)

## Contact

Any questions or issues feel free to get in touch

[hello@robhayward.co.uk] (hello@robhayward.co.uk)

[@robhayward] (http://www.twitter.com/robhayward)
