//
//  RHTableViewProvider.m
//  RHTableViewProvider
//
//  Created by Rob Hayward on 11/01/2013.
//  Copyright (c) 2013 Rob Hayward. All rights reserved.
//

#import "RHTableViewProvider.h"
#import "RHTableViewProviderCell.h"
#import "RHTableViewProviderSection.h"
#import "NSDictionary+RHTVP.h"
#import "NSIndexPath+RHTVP.h"

NSString *const RHTVPSectionHeader = @"RHTVPSectionHeader";
NSString *const RHTVPSectionFooter = @"RHTVPSectionFooter";
NSString *const RHTVPSectionRows = @"RHTVPSectionRows";
NSString *const RHTVPSectionExpandable = @"RHTVPSectionExpandable";

@interface RHTableViewProvider ()
{
  BOOL _hasSections;
  NSInteger _totalItems;
}
@property (strong, nonatomic) NSMutableArray *content;
@property (assign, nonatomic) BOOL hasScrolledToBottom;

@end

@implementation RHTableViewProvider

+ (UITableView *)tableViewWithFrame:(CGRect)frame style:(UITableViewStyle)style forSuperView:(UIView *)view
{
  UITableView *tableView = [[UITableView alloc] initWithFrame:frame style:style];
  [view addSubview:tableView];
  return tableView;
}

- (id)initWithTableView:(UITableView *)aTableView delegate:(id<RHTableViewProviderDelegate>)aDelegate customise:(BOOL)customise
{
  self = [super init];
  if (self)
  {
    NSParameterAssert(aTableView);
    NSParameterAssert(aDelegate);
    [self setTableView:aTableView];
    [self setDelegate:aDelegate];
    [self setShouldDrawCustomViews:customise];
    [self setup];
  }
  return self;
}

#pragma mark - Getters

- (id)objectAtIndexPath:(NSIndexPath *)indexPath
{
  if (self.fetchedResultsController)
  {
    return [self.fetchedResultsController objectAtIndexPath:indexPath];
  }
  
  if (![self hasContent] ) { return nil; }
  
  NSDictionary *section = [self.content objectAtIndex:indexPath.section];
  NSArray *rows = [section valueForKey:RHTVPSectionRows];
  return [rows objectAtIndex:indexPath.row];
}

- (id)objectForSectionAtIndex:(NSInteger)index header:(BOOL)header
{
  if (self.fetchedResultsController)
  {
    if (header)
    {
      id <NSFetchedResultsSectionInfo> sectionInfo = [[self.fetchedResultsController sections] objectAtIndex:index];
      return [sectionInfo name];
    }
    return nil;
  }
  
  NSDictionary *section = [self.content objectAtIndex:index];
  if (header) {
    return [section objectForKeyNotNull:RHTVPSectionHeader];
  }
  return [section objectForKeyNotNull:RHTVPSectionFooter];
}

- (NSIndexPath *)indexPathOfFirstRow
{
  return [NSIndexPath indexPathForItem:0 inSection:0];
}

- (NSIndexPath *)indexPathOfLastRow
{
  if (_indexPathOfLastRow != nil) {
    return _indexPathOfLastRow;
  }
  
  NSInteger section = [self.content count];
  if (section > 0) { section -= 1; }
  NSInteger count = [[[self.content objectAtIndex:section] objectForKey:RHTVPSectionRows] count];
  if (count > 1) { count -= 1; }
  self.indexPathOfLastRow = [NSIndexPath indexPathForItem:count inSection:section];
  return _indexPathOfLastRow;
}

- (NSMutableIndexSet *)expandableSections
{
  if (!_expandableSections) {
    self.expandableSections = [[NSMutableIndexSet alloc] init];
  }
  return _expandableSections;
}

- (NSMutableIndexSet *)expandedSections
{
  if (!_expandedSections) {
    self.expandedSections = [[NSMutableIndexSet alloc] init];
  }
  return _expandedSections;
}

- (NSMutableArray *)expandedCells
{
  if (!_expandedCells) {
    self.expandedCells = [NSMutableArray arrayWithCapacity:0];
  }
  return _expandedCells;
}

- (BOOL)isIndexPathFirstRowOfSection:(NSIndexPath *)indexPath
{
  if (indexPath.row == 0) { return YES; }
  return NO;
}

- (BOOL)isIndexPathLastRowOfSection:(NSIndexPath *)indexPath
{
  if (self.fetchedResultsController) {
    id <NSFetchedResultsSectionInfo> sectionInfo = [[self.fetchedResultsController sections] objectAtIndex:indexPath.section];
    NSInteger count = [sectionInfo numberOfObjects];
    if (indexPath.row == (count - 1)) { return YES; }
    return NO;
  }
  
  NSInteger count = [[[self.content objectAtIndex:indexPath.section] objectForKey:RHTVPSectionRows] count];
  if (indexPath.row == (count - 1)) { return YES; }
  return NO;
}

- (BOOL)isCellSingleInSectionForIndexPath:(NSIndexPath *)indexPath
{
  if (self.fetchedResultsController) {
    id <NSFetchedResultsSectionInfo> sectionInfo = [[self.fetchedResultsController sections] objectAtIndex:indexPath.section];
    NSInteger count = [sectionInfo numberOfObjects];
    if (count > 1) {
      return NO;
    }
    return YES;
  }
  
  NSInteger count = [[[self.content objectAtIndex:indexPath.section] objectForKey:RHTVPSectionRows] count];
  if (count > 1) {
    return NO;
  }
  return YES;
}

- (CGFloat)cellWidth
{
  CGFloat fullWidth = self.tableView.bounds.size.width;
  
  if (!self.tableView.style == UITableViewStyleGrouped) { return fullWidth; }
  else
  {
    CGFloat margins = ((self.tableView.bounds.size.width * GROUPED_CELL_WIDTH_MULTIPLIER) * 2);
    return (fullWidth - margins);
  }
}

- (NSMutableArray *)contentWithSections:(NSArray *)theContent
{
  NSMutableArray *mutable = [NSMutableArray arrayWithCapacity:0];
  [mutable addObject:@{RHTVPSectionRows: [theContent mutableCopy]}];
  return mutable;
}

- (BOOL)isIndexPathSelected:(NSIndexPath *)indexPath
{
  NSIndexPath *selectedIndexPath = [_tableView indexPathForSelectedRow];
  if (selectedIndexPath)
  {
    if ([selectedIndexPath isEqualToIndexPath:indexPath])
    {
      return YES;
    }
    
  }
  return NO;
}

#pragma mark - Expanding Cells

- (BOOL)tableViewCellCanExpand:(NSIndexPath *)indexPath
{
  if (!_shouldDrawCustomViews) { return NO; }
  
  return [[self tableCellClassForRowAtIndexPath:indexPath] expandable];
}

- (BOOL)tableViewCellIsExpanded:(NSIndexPath *)indexPath
{
  for (NSIndexPath *ip in self.expandedCells)
  {
    if ([ip isEqualToIndexPath:indexPath]) {
      return YES;
    }
  }
  return NO;
}
                          
- (void)tableViewCellToRemoveFromExpandedCells:(NSIndexPath *)indexPath
{
  NSIndexPath *marked = nil;
  for (NSIndexPath *ip in self.expandedCells)
  {
    if ([ip isEqualToIndexPath:indexPath])
    {
      marked = ip;
      break;
    }
  }
  if (marked) {
    [self.expandedCells removeObject:marked];
  }
}

#pragma mark - Expanding Sections

- (void)updateExpandableSections
{
  NSInteger sectionCount = [_content count];
  for (NSInteger i = 0; i < sectionCount; i++)
  {
    BOOL expandable = [[[_content objectAtIndex:i] valueForKey:RHTVPSectionExpandable] boolValue];
    if (expandable) {
      [self.expandableSections addIndex:i];
    } else {
      [self.expandableSections removeIndex:i];
    }
  }
}

- (BOOL)tableViewSectionCanExpand:(NSInteger)section
{
  return [self.expandableSections containsIndex:section];
}

- (BOOL)tableViewSectionIsExpanded:(NSInteger)section
{
  return [self.expandedSections containsIndex:section];
}

#pragma mark - Delete

- (void)deleteObjectAtIndexPath:(NSIndexPath *)indexPath
{
  if (self.fetchedResultsController)
  {
    id object = [self objectAtIndexPath:indexPath];
    [self.context deleteObject:object];
    return;
  }
  NSMutableDictionary *section =  [NSMutableDictionary dictionaryWithDictionary:[self.content objectAtIndex:indexPath.section]];
  NSMutableArray *rows = [NSMutableArray arrayWithArray:[section valueForKey:RHTVPSectionRows]];
  id object = [rows objectAtIndex:indexPath.row];
  if (object) {
    [rows removeObjectAtIndex:indexPath.row];
  }
  [section setValue:rows forKey:RHTVPSectionRows];
  [self.content replaceObjectAtIndex:indexPath.section withObject:section];
}

#pragma mark - Setters

- (void)setContent:(NSArray *)theContent withSections:(BOOL)sections
{
  self.fetchedResultsController = nil;
  self.fetchRequest = nil;
  _hasSections = sections;
  
  if (theContent == nil)
  {
    self.content = [NSMutableArray arrayWithCapacity:0];
  }
  else
  {
    if (_hasSections) {
      self.content = [theContent mutableCopy];
    } else {
      self.content = [self contentWithSections:theContent];
    }
  }
  [self reload];
}

- (void)setContentWithFetchRequest:(NSFetchRequest *)aFetchRequest context:(NSManagedObjectContext *)aContext
{
  [self setContentWithFetchRequest:aFetchRequest sectionKeyPath:nil context:aContext];
}

- (void)setContentWithFetchRequest:(NSFetchRequest *)aFetchRequest sectionKeyPath:(NSString *)keyPath context:(NSManagedObjectContext *)aContext
{
  self.fetchRequest = aFetchRequest;
  self.context = aContext;
  self.fetchedResultsController = nil;
  self.sectionKeyPath = keyPath;
  if (self.sectionKeyPath) { _hasSections = YES; } else { _hasSections = NO; }
  
  NSError *error = nil;
  
	if (![[self fetchedResultsController] performFetch:&error]) {
		NSLog(@"Unresolved error %@, %@", error, [error userInfo]);
		exit(-1);
	}
  
  [self reload];
}

#pragma mark - Total Items

- (void)updateTotalItems
{
  _totalItems = [self totalItemsCount];
}

- (NSInteger)totalItemsCount
{
  if (self.fetchedResultsController)
  {
    return [[self.fetchedResultsController fetchedObjects] count];
  }
  
  NSInteger count = 0;
  
  for (NSDictionary *section in self.content) {
    count += [[section valueForKey:RHTVPSectionRows] count];
  }
  
  return count;
}

- (BOOL)hasContent {
  
  if (self.fetchedResultsController)
  {
    if ([[self.fetchedResultsController fetchedObjects] count] > 0) {
      return YES;
    }
    return NO;
  }
  
  if (self.content) {
    if ([self.content count] > 0) {
      return YES;
    }
  }
  
  return NO;
}

#pragma mark - Reload

- (void)reload
{
  self.indexPathOfLastRow = nil;
  
  BOOL hasContent = [self hasContent];
  if (hasContent)
  {
    [self updateExpandableSections];
    [self displayWithData];
    [self updateTotalItems];
    [self.tableView reloadData];
  } else {
    [self displayWithoutData];
    [self.tableView reloadData];
  }
}

- (void)reloadVisibleCells
{
  NSArray *visibleCells = self.tableView.visibleCells;
  for (RHTableViewProviderCell *cell in visibleCells)
  {
    id object = [self objectAtIndexPath:cell.indexPath];
    [cell populateWithObject:object];
  }
}

#pragma mark - Display State

- (void)displayWithData
{
  [self.emptyView setHidden:YES];
  [self.tableView setHidden:NO];
  if ([_delegate respondsToSelector:@selector(RHTableViewProviderDidUpdateWithContent:)]) {
    [_delegate RHTableViewProviderDidUpdateWithContent:self];
  }
}

- (void)displayWithoutData
{
  [self.tableView setHidden:YES];
  [self.emptyView setHidden:NO];
  if ([_delegate respondsToSelector:@selector(RHTableViewProviderDidUpdateWithoutContent:)]) {
    [_delegate RHTableViewProviderDidUpdateWithoutContent:self];
  }
}

#pragma mark - UITableViewDelegate

- (BOOL)tableView:(UITableView *)tableView shouldShowMenuForRowAtIndexPath:(NSIndexPath *)indexPath
{
  // Subclass the provider to implement this functionality
  return NO;
}

#pragma mark - Sections

- (NSInteger)tableView:(UITableView *)tableView numberOfRowsInSection:(NSInteger)section
{
  NSInteger count = 0;
  if (self.fetchedResultsController)
  {
    id sectionInfo = [[self.fetchedResultsController sections] objectAtIndex:section];
    count = [sectionInfo numberOfObjects];
  }
  else
  {
    NSMutableDictionary *sectionContent = [self.content objectAtIndex:section];
    count = [[sectionContent valueForKey:RHTVPSectionRows] count];
  }
  
  if ([self tableViewSectionCanExpand:section])
  {
    if (![self tableViewSectionIsExpanded:section])
    {
      return 1;
    }
  }
  
  return count;
}

- (NSInteger)numberOfSectionsInTableView:(UITableView *)tableView
{
  NSInteger count = 0;
  if (self.fetchedResultsController)
  {
    count = [[self.fetchedResultsController sections] count];
    return count;
  }
  count = [self.content count];
  return count;
}

- (CGFloat)heightForSectionAtIndex:(NSInteger)index header:(BOOL)header
{
  CGFloat height = 0.0f;
  
  if (_hasSections) { height = self.defaultSectionHeight; }
  
  Class viewClass = [self tableSectionViewClassForSection:index header:header];
  if (viewClass)
  { return [viewClass height]; }
  else
  {
    id object = [self objectForSectionAtIndex:index header:header];
    if (!object) { height = 0.0f; }
  }
  return height;
}

- (CGFloat)tableView:(UITableView *)tableView heightForHeaderInSection:(NSInteger)section
{
  return [self heightForSectionAtIndex:section header:YES];
}

- (CGFloat)tableView:(UITableView *)tableView heightForFooterInSection:(NSInteger)section
{
  return [self heightForSectionAtIndex:section header:NO];
}

- (NSArray *)sectionIndexTitlesForTableView:(UITableView *)tableView
{
  return self.sectionIndexTitles;
}

- (NSString *)tableView:(UITableView *)tableView titleForHeaderInSection:(NSInteger)section
{
  if (self.fetchedResultsController) { return nil; } // TODO.. Implement!
  return [self objectForSectionAtIndex:section header:YES];
}

- (NSString *)tableView:(UITableView *)tableView titleForFooterInSection:(NSInteger)section
{
  if (self.fetchedResultsController) { return nil; } // TODO.. Implement!
  return [self objectForSectionAtIndex:section header:NO];
}

- (UIView *)tableView:(UITableView *)tableView viewForHeaderInSection:(NSInteger)section
{
  Class viewClass = [self tableSectionViewClassForSection:section header:YES];
  if (!viewClass) {
    return nil;
  }
  RHTableViewProviderSection *view = (RHTableViewProviderSection *)[[viewClass alloc] initWithFrame:CGRectMake(0, 0, self.tableView.bounds.size.width, [viewClass height])];
  
  id object = [self objectForSectionAtIndex:section header:YES];
  [view populateWithObject:object];
  
  return (UIView *)view;
}

- (UIView *)tableView:(UITableView *)tableView viewForFooterInSection:(NSInteger)section
{
  Class viewClass = [self tableSectionViewClassForSection:section header:NO];
  if (!viewClass) {
    return nil;
  }
  RHTableViewProviderSection *view = (RHTableViewProviderSection *)[[viewClass alloc] initWithFrame:CGRectMake(0, 0, self.tableView.bounds.size.width, [viewClass height])];
  
  id object = [self objectForSectionAtIndex:section header:NO];
  [view populateWithObject:object];
  
  return (UIView *)view;
}

- (void)tableView:(UITableView *)tableView willDisplayHeaderView:(UIView *)view forSection:(NSInteger)section
{
  if (_shouldDrawCustomViews) { [(RHTableViewProviderSection *)view willDisplay]; }
}

- (void)tableView:(UITableView *)tableView willDisplayFooterView:(UIView *)view forSection:(NSInteger)section
{
  if (_shouldDrawCustomViews) { [(RHTableViewProviderSection *)view willDisplay]; }
}

- (void)tableView:(UITableView *)tableView didEndDisplayingHeaderView:(UIView *)view forSection:(NSInteger)section
{
  if (_shouldDrawCustomViews) { [(RHTableViewProviderSection *)view didEndDisplay]; }
}

- (void)tableView:(UITableView *)tableView didEndDisplayingFooterView:(UIView *)view forSection:(NSInteger)section
{
  if (_shouldDrawCustomViews) { [(RHTableViewProviderSection *)view didEndDisplay]; }
}

#pragma mark - Rows

- (CGFloat)tableView:(UITableView *)tableView heightForRowAtIndexPath:(NSIndexPath *)indexPath
{
  CGFloat height = 44.0f;
  if (_shouldDrawCustomViews)
  {
    Class cellClass = [self tableCellClassForRowAtIndexPath:indexPath];
    id object = [self objectAtIndexPath:indexPath];
    
    if ( [object conformsToProtocol:@protocol(RHTableViewProviderContent)] )
    {
      height = [object contentHeight];
      if ([self tableViewCellCanExpand:indexPath])
      {
        if ([self tableViewCellIsExpanded:indexPath]) {
          height = [object contentHeightWhenExpanded];
        }
      }
      height += ([cellClass margins] * 2);
    }
    else
    {
      height = [cellClass height];
      if ([self tableViewCellCanExpand:indexPath])
      {
        if ([self tableViewCellIsExpanded:indexPath]) {
          height = [[self tableCellClassForRowAtIndexPath:indexPath] heightForSelectedState];
        }
      }
    }
  }
  return MAX(height, MINIMUM_CELL_HEIGHT);
}

#pragma mark - Cells

- (UITableViewCell *)tableView:(UITableView *)tableView cellForRowAtIndexPath:(NSIndexPath *)indexPath
{
  id object = [self objectAtIndexPath:indexPath];
  Class cellClass = [self tableCellClassForRowAtIndexPath:indexPath];
  NSString *identifier = NSStringFromClass(cellClass);
  UITableViewCell *cell = [tableView dequeueReusableCellWithIdentifier:identifier];
  if (cell == nil)
  {
    cell = [[cellClass alloc] initWithStyle:UITableViewCellStyleDefault reuseIdentifier:identifier];
    if ([cell isKindOfClass:[RHTableViewProviderCell class]])
    {
      [(RHTableViewProviderCell *)cell setProvider:self];
      [(RHTableViewProviderCell *)cell setParentTableView:_tableView];
    }
  }
  if ([cell isKindOfClass:[RHTableViewProviderCell class]])
  {
    if      ([self isCellSingleInSectionForIndexPath:indexPath]) { [(RHTableViewProviderCell *)cell setCellType:RHTableViewProviderCellTypeSingle]; }
    else if ([self isIndexPathFirstRowOfSection:indexPath]) { [(RHTableViewProviderCell *)cell setCellType:RHTableViewProviderCellTypeFirst]; }
    else if ([self isIndexPathLastRowOfSection:indexPath]) { [(RHTableViewProviderCell *)cell setCellType:RHTableViewProviderCellTypeLast]; }
    else {  [(RHTableViewProviderCell *)cell setCellType:RHTableViewProviderCellTypeMiddle]; }
    
    [(RHTableViewProviderCell *)cell setIndexPath:indexPath];
    [(RHTableViewProviderCell *)cell populateWithObject:object];
  }
  else
  {
    cell.textLabel.text = object;
  }
  return cell;
}

- (void)tableView:(UITableView *)tableView willDisplayCell:(UITableViewCell *)cell forRowAtIndexPath:(NSIndexPath *)indexPath
{
  if ([cell isKindOfClass:[RHTableViewProviderCell class]])
  {
    [(RHTableViewProviderCell *)cell willDisplay];
  }
}

- (NSInteger)tableView:(UITableView *)tableView indentationLevelForRowAtIndexPath:(NSIndexPath *)indexPath
{
  // Subclass the provider to implement this functionality
  return 0;
}

#pragma mark - Row Selection

- (void)tableView:(UITableView *)tableView didSelectRowAtIndexPath:(NSIndexPath *)indexPath
{
  
  if ([self tableViewCellCanExpand:indexPath])
  {
    [tableView beginUpdates];
    [tableView endUpdates];
  }
  
  
  if ([self tableViewSectionCanExpand:indexPath.section])
  {
    if (!indexPath.row)
    {
      // only first row toggles exapand/collapse
      [tableView deselectRowAtIndexPath:indexPath animated:YES];
      [self tableView:tableView didDeselectRowAtIndexPath:indexPath];
      
      NSInteger section = indexPath.section;
      BOOL expanded = [self tableViewSectionIsExpanded:section];
      NSInteger rows = 0;
      NSMutableArray *tmpArray = [NSMutableArray arrayWithCapacity:0];
      
      if (expanded)
      {
        rows = [self tableView:tableView numberOfRowsInSection:section];
        [self.expandedSections removeIndex:section];
        
      }
      else
      {
        [self.expandedSections addIndex:section];
        rows = [self tableView:tableView numberOfRowsInSection:section];
      }
      
      for (NSInteger i = 1; i < rows; i++)
      {
        NSIndexPath *tmpIndexPath = [NSIndexPath indexPathForRow:i inSection:section];
        [tmpArray addObject:tmpIndexPath];
      }
      
      if (expanded)
      {
        [tableView deleteRowsAtIndexPaths:tmpArray withRowAnimation:UITableViewRowAnimationTop];
        
      }
      else
      {
        [tableView insertRowsAtIndexPaths:tmpArray withRowAnimation:UITableViewRowAnimationTop];
        
      }
    }
  }
  
  if (_shouldClearCellSelection)
  {
    [tableView deselectRowAtIndexPath:indexPath animated:NO];
  }
  
  if ([self.delegate respondsToSelector:@selector(RHTableViewProvider:didSelectRowAtIndexPath:)])
  {
    [self.delegate RHTableViewProvider:self didSelectRowAtIndexPath:indexPath];
  }
}

- (NSIndexPath *)tableView:(UITableView *)tableView willSelectRowAtIndexPath:(NSIndexPath *)indexPath
{
  // Subclass the provider to implement this functionality
  if ([self tableViewCellCanExpand:indexPath])
  {
    BOOL expanded = [self tableViewCellIsExpanded:indexPath];
    if (expanded)
    {
      [self tableViewCellToRemoveFromExpandedCells:indexPath];
    }
    else
    {
      [self.expandedCells addObject:indexPath];
    }
  }
  return indexPath;
}

- (NSIndexPath *)tableView:(UITableView *)tableView willDeselectRowAtIndexPath:(NSIndexPath *)indexPath
{
  // Subclass the provider to implement this functionality
  return indexPath;
}

- (void)tableView:(UITableView *)tableView didDeselectRowAtIndexPath:(NSIndexPath *)indexPath
{
  [tableView beginUpdates];
  [tableView endUpdates];
}

- (void)tableView:(UITableView *)tableView accessoryButtonTappedForRowWithIndexPath:(NSIndexPath *)indexPath
{
  [self tableView:tableView didSelectRowAtIndexPath:indexPath];
}

#pragma mark - Row Highlight

- (BOOL)tableView:(UITableView *)tableView shouldHighlightRowAtIndexPath:(NSIndexPath *)indexPath
{
  return YES;
}

- (void)tableView:(UITableView *)tableView didHighlightRowAtIndexPath:(NSIndexPath *)indexPath
{
  // Subclass the provider to implement this functionality
}

- (void)tableView:(UITableView *)tableView didUnhighlightRowAtIndexPath:(NSIndexPath *)indexPath
{
  // Subclass the provider to implement this functionality
  
}

#pragma mark - Custom Views

- (Class)tableCellClassForRowAtIndexPath:(NSIndexPath *)indexPath
{
  if (!_shouldDrawCustomViews) { return [UITableViewCell class]; }
  
  NSString *name = nil;
  if ([self.delegate respondsToSelector:@selector(RHTableViewProvider:cellClassForRowAtIndexPath:)]) {
    name = [self.delegate RHTableViewProvider:self cellClassForRowAtIndexPath:indexPath];
  }
  if (name == nil)
  {
    if (self.tableView.style == UITableViewStyleGrouped) {
      name = self.defaultGroupedCellClassName;
    } else { name = self.defaultCellClassName; }
  }
  return NSClassFromString(name);
}

- (Class)tableSectionViewClassForSection:(NSInteger)section header:(BOOL)header
{
  if (!_shouldDrawCustomViews) { return nil; }
  
  id object = [self objectForSectionAtIndex:section header:header];
  if (!object) {
    return nil;
  }
  
  NSString *name = nil;
  
  if (header)
  {
    if (self.tableView.style == UITableViewStylePlain) { name = self.defaultSectionHeaderViewClassName; }
    else { name = name = self.defaultGroupedSectionHeaderViewClassName; }
    
    if ([self.delegate respondsToSelector:@selector(RHTableViewProvider:sectionHeaderViewClassForSection:)])
    {
      name = [self.delegate RHTableViewProvider:self sectionHeaderViewClassForSection:section];
    }
  }
  else
  {
    if (self.tableView.style == UITableViewStylePlain) { name = self.defaultSectionFooterViewClassName; }
    else { name = name = self.defaultGroupedSectionFooterViewClassName; }
    
    if ([self.delegate respondsToSelector:@selector(RHTableViewProvider:sectionFooterViewClassForSection:)]) {
      name = [self.delegate RHTableViewProvider:self sectionFooterViewClassForSection:section];
    }
  }
  
  if (name != nil) {
    return NSClassFromString(name);
  }
  
  return nil;
}

#pragma mark - UIScrollViewDelegate

- (void)scrollViewDidScroll:(UIScrollView *)scrollView
{
  [self tableViewScrolledToBottomCheck];
}

#pragma mark - Scrolls to bottom

- (void)tableViewScrolledToBottomCheck
{
  CGFloat height = _tableView.bounds.size.height;
  CGFloat size   = _tableView.contentSize.height;
  CGFloat offset = _tableView.contentOffset.y;
  CGFloat test   = size - height;
  
  if (offset < test) {
    _hasScrolledToBottom = NO;
  }
  else if (offset >= test) {
    
    if (_hasScrolledToBottom == YES) {
      return;
    }
    
    _hasScrolledToBottom = YES;
    if ([self.delegate respondsToSelector:@selector(RHTableViewProviderDidScrollToBottom:)]) {
      [self.delegate RHTableViewProviderDidScrollToBottom:self];
    }
  }
}

#pragma mark - Core Data

- (NSFetchedResultsController *)fetchedResultsController
{
  if (self.fetchRequest == nil) { return nil; }
  if (_fetchedResultsController != nil) { return _fetchedResultsController; }
  
  NSString *cacheName = nil;
  NSFetchedResultsController *theFetchedResultsController = [[NSFetchedResultsController alloc] initWithFetchRequest:self.fetchRequest managedObjectContext:self.context sectionNameKeyPath:self.sectionKeyPath cacheName:cacheName];
  self.fetchedResultsController = theFetchedResultsController;
  self.fetchedResultsController.delegate = self;
  
  return _fetchedResultsController;
}

#pragma mark - NSFetchedResultsControllerDelegate

- (void)controller:(NSFetchedResultsController *)controller didChangeObject:(id)anObject atIndexPath:(NSIndexPath *)indexPath forChangeType:(NSFetchedResultsChangeType)type newIndexPath:(NSIndexPath *)newIndexPath
{
  // Subclass the provider to implement this functionality
}

- (void)controller:(NSFetchedResultsController *)controller didChangeSection:(id<NSFetchedResultsSectionInfo>)sectionInfo atIndex:(NSUInteger)sectionIndex forChangeType:(NSFetchedResultsChangeType)type
{
  // Subclass the provider to implement this functionality
}

#pragma mark - Setup

- (void)setup
{
  [self setupTableView];
  [self setupDefaults];
}

- (void)setupTableView
{
  [_tableView setDataSource:self];
  [_tableView setDelegate:self];
  [_tableView setTableFooterView:[UIView new]];
  if (_shouldDrawCustomViews) {
    [_tableView setSeparatorStyle:UITableViewCellSeparatorStyleNone];
  }
}

- (void)setupDefaults
{
  self.defaultSectionHeight = 30.0f;
  self.defaultCellClassName = @"RHTableViewProviderCellDefault";
  self.defaultGroupedCellClassName = @"RHTableViewProviderCellGroupedDefault";
  self.defaultSectionHeaderViewClassName = @"RHTableViewProviderSectionViewDefault";
  self.defaultSectionFooterViewClassName = @"RHTableViewProviderSectionViewDefault";
  self.defaultGroupedSectionHeaderViewClassName = @"RHTableViewProviderSectionViewGroupedDefault";
  self.defaultGroupedSectionFooterViewClassName = @"RHTableViewProviderSectionViewGroupedDefault";
}

@end
